// server_env.hpp
//
// read the server.env file and provide a class to access the variables

#ifndef __SERVER_ENV_HPP
#define __SERVER_ENV_HPP

#include <string>
#include <map>
#include <memory>
#include <vector>
#include <set>
#include <optional>

#include "utils/execute.hpp"

namespace dropshell
{

    struct UserConfig
    {
        std::string user;
        std::string dir;
    };

    // ------------------------------------------------------------------------------------------------

    // reads path / server.env and provides a class to access the variables.
    // each env file is required to have the following variables:
    // SSH_HOST
    // SSH_UNPRIVILEGED_USER
    // SSH_PORT
    // the following replacements are made in the values:
    // ${USER} -> the username of the user running dropshell
    class ServerConfig
    {
    public:
        ServerConfig(const std::string &server_name);

        bool is_valid() const { return mValid; }


        // trivial getters.
        const std::map<std::string, std::string> &get_variables() const { return mVariables; }
        std::string get_variable(const std::string &name) const;

        // ------------------------------------------------------------------------------------------------
        // getters
        // ------------------------------------------------------------------------------------------------
        std::string get_SSH_HOST() const;
        std::string get_SSH_PORT() const;
        std::vector<UserConfig> get_users() const;
        std::string get_user_dir(const std::string &user) const;
        std::string get_server_name() const;
        std::string get_user_for_service(const std::string &service) const;

        sSSHInfo get_SSH_INFO(std::string user) const;

        // server capabilities
        bool hasRootUser() const;
        bool hasDocker() const;
        bool hasRootDocker() const;

        // helper functions
        bool hasUser(const std::string &user) const;

    public:
        bool check_remote_dir_exists(const std::string &dir_path, std::string user) const;
        bool check_remote_file_exists(const std::string &file_path, std::string user) const;
        bool check_remote_items_exist(const std::vector<std::string> &file_paths, std::string user) const;

        bool remove_remote_dir(const std::string &dir_path, bool silent, std::string user) const;

        bool run_remote_template_command(const std::string &service_name, const std::string &command,
                                         std::vector<std::string> args, bool silent, std::map<std::string, std::string> extra_env_vars) const;

        bool run_remote_template_command_and_capture_output(const std::string &service_name, const std::string &command,
                                                            std::vector<std::string> args, std::string &output, bool silent, std::map<std::string, std::string> extra_env_vars) const;

    private:
        std::optional<sCommand> construct_standard_template_run_cmd(const std::string &service_name, const std::string &command, const std::vector<std::string> args, const bool silent) const;

    private:
        std::string mServerName;
        std::map<std::string, std::string> mVariables;
        std::vector<UserConfig> mUsers;
        bool mValid;
    }; // class ServerConfig



    std::vector<ServerConfig> get_configured_servers();

    std::string get_user_for_service(const std::string &server, const std::string &service);

    bool create_server(const std::string &server_name);

    void get_all_used_commands(std::set<std::string> &commands);

    bool server_exists(const std::string &server_name);

} // namespace dropshell

#endif // __SERVER_ENV_HPP